import click


def option_repo_config_file(function):
    function = click.option(
        "-c",
        "--config-repo-file",
        default="repos.yml",
        type=click.Path(
            exists=True,
            dir_okay=False,
            resolve_path=True,
        ),
        help="Repositories file",
    )(function)
    return function



import subprocess
from pathlib import Path

import click
from loguru import logger

from .tools import (
    ensure_line,
    git_checkout,
    git_commit_all,
    git_push,
    replace_expression,
)


@click.command()
@click.option(
    "-i",
    "--initial-version",
    type=str,
    required=True,
    help="origin version, which contains the answer copy file.",
    default="12.0",
)
@click.option(
    "-t",
    "--target-version",
    type=str,
    required=True,
    help="target version to create.",
    default="16.0",
)
@click.option(
    "-r",
    "--remote",
    type=str,
    required=True,
    help="Remote name. If the reference repository is"
    " https://github.com/OCA/pos, set 'OCA'.",
    default="grap",
)
@click.option(
    "-u",
    "--copier-url",
    type=str,
    required=True,
    help="Git URL of the copier template to use for the new branch.",
    default="https://github.com/grap/oca-addons-repo-template-v16",
)
@click.option("-f", "--force", is_flag=True, show_default=True, default=False)
@click.pass_context
def create_branch(ctx, initial_version, target_version, remote, copier_url, force):
    """Create a new branch for Odoo addons"""

    # recover Copier answers from initial version
    git_checkout(initial_version)
    copier_file = Path(".copier-answers.yml").resolve()
    if copier_file.exists():
        with copier_file.open(mode="r") as file:
            copier_content = file.read()
    else:
        copier_content = ""

    # create empty git branch
    git_checkout(target_version, mode="create_orphan")
    if copier_content:
        with copier_file.open(mode="w") as file:
            copier_content = file.write(copier_content)

        git_commit_all(
            "[INIT] initialize .copier-answers.yml file"
            f" with file present in {initial_version}."
        )

    # Adapt copier-answers.yml file to recent GRAP convention
    logger.info("Adapt copier answers file to new conventions ...")
    replace_expression(copier_file, "odoo_real_version: 12.0", "")
    ensure_line(copier_file, "convert_readme_fragments_to_markdown: true")
    ensure_line(copier_file, "odoo_test_flavor: Odoo")
    ensure_line(copier_file, "use_ruff: true")

    git_commit_all(
        f"[INIT] Update .copier-answers.yml answers"
        f" with {remote} conventions for {target_version} version."
    )

    subprocess.check_call(["copier", "copy", copier_url, ".", "--trust"])

    git_commit_all("[INIT] Initialize repo with copier template.", verify=True)

    git_push(remote, target_version, force)



import subprocess
from datetime import datetime
from pathlib import Path

import click
from git import Repo
from github import Auth, Github, GithubException
from loguru import logger

from .tools import execute, git_push, git_switch


@click.command()
@click.option("-i", "--initial-version", type=str, required=True, default="12.0")
@click.option("-t", "--target-version", type=str, required=True, default="16.0")
@click.option("-r", "--distant-remote", type=str, required=True, default="grap")
@click.option("-l", "--local-remote", type=str, required=True)
@click.option("-m", "--modules", type=str)
@click.option(
    "-g",
    "--github-token-file",
    type=click.Path(
        exists=True,
        dir_okay=False,
        resolve_path=True,
    ),
)
@click.option("-f", "--force", is_flag=True, show_default=True, default=False)
@click.pass_context
def migrate(
    ctx,
    initial_version,
    target_version,
    distant_remote,
    local_remote,
    modules,
    github_token_file,
    force,
):
    """Migrate modules from a version X to version Y"""
    current_dir = Path("./").resolve()
    repo = Repo(str(current_dir))

    logger.info(f"Read github token in '{github_token_file}' ...")
    github_token_file = Path(github_token_file)
    with github_token_file.open(mode="r") as file:
        token = file.read()

    # Switch and pull last revision of initial version
    git_switch(repo, distant_remote, initial_version, do_pull=True)

    # Get modules, if not defined
    modules = modules and modules.split(",") or []
    modules_path = []

    if not modules:
        for directory in [x for x in current_dir.iterdir() if x.is_dir()]:
            if (directory / "__manifest__.py").exists():
                modules_path.append(directory)
    else:
        for module in modules:
            module_path = (Path("./") / module).resolve()
            if module_path.exists():
                modules_path.append(module_path)
            else:
                logger.warning(f"{module} not found in the branch {initial_version}")

    # Switch and pull last revision of target version
    git_switch(repo, distant_remote, target_version, do_pull=True)

    for module_path in modules_path:
        module_name = module_path.name
        patch_path = (
            Path("/tmp/") / datetime.utcnow().strftime("%Y-%m-%d__%H-%M-%S-%f")
        ).resolve()
        branch_name = f"{target_version}-mig-{module_name}"

        logger.info(f"Migrating {module_name} ...")
        git_switch(repo, distant_remote, target_version)
        if branch_name in repo.branches:
            if force:
                value = "D"
            else:
                value = click.prompt(
                    f"Branch {branch_name} exist:"
                    " [s]kip migration or [d]rop branch ?",
                    type=str,
                    default="s",
                )
            if value in ["s", "S"]:
                logger.info(f"Skipping migration of {module_name}")
                continue
            elif value in ["d", "D"]:
                logger.warning(f"Dropping branch {branch_name}")
                execute(["git", "branch", "-D", branch_name])
            else:
                logger.critical("Incorrect answer, exiting...")
                return

        logger.info(f"Creating new branch {branch_name} ...")
        execute(["git", "checkout", "-b", branch_name])

        logger.info(f"Creating patches in {patch_path} ...")
        patch_path.mkdir(parents=True, exist_ok=True)
        execute(
            [
                "git",
                "format-patch",
                "--keep-subject",
                f"--output-directory={patch_path}",
                f"{distant_remote}/{target_version}"
                f"..{distant_remote}/{initial_version}",
                "--",
                f"{module_name}",
            ]
        )
        patch_files = sorted([x for x in patch_path.iterdir()])

        subprocess.check_call(["pre-commit", "uninstall"])

        logger.info(f"Applying {len(patch_files)} patches ...")
        last_summary = ""
        for i, patch_file in enumerate(patch_files, 1):
            logger.info(
                f"{i:>02}/{len(patch_files):>02}"
                f" Applying patch {patch_file.name} ..."
            )
            execute(["git", "am", "-3", "--keep", str(patch_file)])
            current_summary = repo.head.commit.summary

            clean_current_summary = (
                current_summary.replace("fixup! ", "")
                .replace("squash! ", "")
                .replace("amend! ", "")
            )
            clean_last_summary = (
                last_summary.replace("fixup! ", "")
                .replace("squash! ", "")
                .replace("amend! ", "")
            )

            if clean_current_summary == clean_last_summary:
                logger.info(
                    "Merging the following commits:\n"
                    f"- commit 1: '{last_summary}'\n"
                    f"- commit 2: '{current_summary}'"
                )
                execute(["git", "reset", "--soft", "HEAD~1"])
                execute(
                    [
                        "git",
                        "commit",
                        "-a",
                        "--amend",
                        "--no-edit",
                        "--allow-empty",
                    ]
                )
            last_summary = current_summary

        logger.info("Call pre-commit ...")
        try:
            execute(["pre-commit", "run", "-a"])
        except Exception:
            try:
                execute(["pre-commit", "run", "-a"])
            except Exception:
                logger.warning("Some pre-commit stuff should be fixed manually.")

        logger.info("Add and commit pre-commit changes in no-verify mode ...")
        execute(["git", "add", "."])
        execute(
            [
                "git",
                "commit",
                "--no-verify",
                "--message",
                f"[IMP] {module_name}: pre-commit stuff",
            ]
        )

        logger.info("Call odoo-module-migrate ...")
        execute(
            [
                "odoo-module-migrate",
                f"--init-version-name={initial_version}",
                f"--target-version-name={target_version}",
                "--no-commit",
                "--no-pre-commit",
                f"--modules={module_name}",
            ]
        )

        logger.info("Add and commit migration changes in no-verify mode ...")
        execute(["git", "add", "."])
        execute(
            [
                "git",
                "commit",
                "--no-verify",
                "--message",
                f"[MIG] {module_name}: Migration to {target_version}"
                f" (from {initial_version})",
            ]
        )

        git_push(local_remote, branch_name)

        logger.info("Creating Pull Request ...")
        gh = Github(auth=Auth.Token(token))
        gh_repo = gh.get_repo(f"{distant_remote}/{current_dir.name}")
        try:
            pull_request = gh_repo.create_pull(
                base=target_version,
                head=f"{local_remote}:{branch_name}",
                title=f"[{target_version}][MIG] {module_name}:"
                f" Migration to {target_version} (from {initial_version})",
                draft=True,
            )
            logger.info(f"Pull request available on {pull_request.html_url}")
        except GithubException as e:
            message_text = "\n- ".join(
                [x.get("message", "") for x in e.data.get("errors", [])]
            )
            logger.error(
                f"Unable to create the pull request on github."
                f" Status: {e.status}"
                f"\n- {message_text}"
            )

    logger.info("End of the script !")



import re
from pathlib import Path

import click
import yaml
from git import Repo
from github import Auth, Github
from loguru import logger

from odoo_tools_grap.cli.click_options import option_repo_config_file

from .tools import git_checkout, git_fetch, git_pull_rebase, git_push


@click.command()
@click.option("-v", "--version", type=str, required=True)
@click.option(
    "-r", "--remotes", type=str, required=True, default=["grap"], multiple=True
)
@click.option("-p", "--projects", type=str, default=[], multiple=True)
@click.option(
    "-g",
    "--github-token-file",
    type=click.Path(
        exists=True,
        dir_okay=False,
        resolve_path=True,
    ),
)
@option_repo_config_file
@click.pass_context
def rebase(ctx, version, remotes, projects, github_token_file, config_repo_file):
    """parse a repo.yml config file, and for each branch,
    rebase the branch and push the result (if it's our branch)
    'Our' is defined with the remote argument.
    """
    logger.info(f"Read github token in '{github_token_file}' ...")
    github_token_file = Path(github_token_file)
    with github_token_file.open(mode="r") as file:
        token = file.read()

    gh = Github(auth=Auth.Token(token))

    config_repo_file = Path(config_repo_file)

    # Compute Addons path
    stream = open(config_repo_file)
    data = yaml.safe_load(stream)

    for addons_path, repo_data in data.items():
        if projects and addons_path.replace("./src/", "") not in projects:
            logger.info(f"[SKIP] {addons_path}.")
            continue
        repo_name = addons_path.split("/")[-1]

        if addons_path == "./src/odoo":
            distant_remote = "odoo"
            # <TODO>
            # Remove this hugly code.
            # In recent grap instance, we base our code on odoo
            # not OCA/OCB
            if version == "12.0":
                distant_remote = "OCA"
                repo_name = "OCB"
            # </TODO>
        elif addons_path == "./src/openupgrade":
            distant_remote = "OCA"
        else:
            distant_remote = addons_path.split("/")[-2]

        path = config_repo_file.cwd() / addons_path
        logger.info(f"Scanning repository '{addons_path}' ...")

        if not path.exists():
            logger.error(
                f"[NOT FOUND] The folder '{addons_path}' has not been found."
                " Maybe you could run:\n"
                f" gitaggregate -c repos.yml -d {addons_path}"
            )
            continue

        repo = Repo(str(path))

        branch_datas = []

        for merge_data in repo_data["merges"]:
            merge_data = merge_data.split(" ")
            pr_remote = merge_data[0]

            if "refs/pull/" in merge_data[1]:
                logger.debug(f"deduce from {merge_data}")
                # We deduce remote and branch name from the pull number
                pr_number = int(re.search(r"pull\/(\d*)", merge_data[1]).group(1))
                gh_repo = gh.get_repo(f"{pr_remote}/{repo_name}")
                try:
                    gh_pr = gh_repo.get_pull(pr_number)
                except Exception:
                    logger.error(
                        f"Unable to get the branch of {pr_remote}/{repo_name}/{pr_number}"
                    )
                    continue
                if gh_pr.is_merged():
                    continue
                merge_data = gh_pr.head.label.replace(":", " ").split(" ")
                branch_datas.append(
                    {
                        "remote": merge_data[0],
                        "branch": merge_data[1],
                        "pr_remote": pr_remote,
                    }
                )

            else:
                branch_datas.append({"remote": merge_data[0], "branch": merge_data[1]})

        # Removing the main branch
        branch_datas = branch_datas[1:]

        for branch_data in branch_datas:
            if branch_data["remote"] not in remotes:
                logger.info(f"[SKIP] {branch_data['remote']} {branch_data['branch']}")
                continue
            logger.info(f"Work on {branch_data['remote']} {branch_data['branch']} ...")
            git_fetch(branch_data["remote"], branch_data["branch"], cwd=path)
            git_checkout(branch_data["branch"], remote=branch_data["remote"], cwd=path)
            last_commit = repo.active_branch.commit.hexsha
            if branch_data.get("pr_remote"):
                remote = branch_data["pr_remote"]
            else:
                remote = distant_remote
            git_pull_rebase(remote, version, cwd=path)
            new_last_commit = repo.active_branch.commit.hexsha
            if last_commit != new_last_commit:
                logger.warning("Commit Changed, pushing new rebased branch ...")
                git_push(
                    branch_data["remote"],
                    branch_data["branch"],
                    force_push=True,
                    cwd=path,
                )



import fileinput
import subprocess
import sys

from loguru import logger


def git_commit_all(message, verify=False, cwd="./"):
    if verify:
        logger.info("Verify All changes ...")
        subprocess.check_call(["pre-commit", "install"], cwd=cwd)
        subprocess.check_call(["git", "add", "."], cwd=cwd)
        subprocess.check_call(["pre-commit", "run", "-a"], cwd=cwd)

    logger.info("Add and commit all changes ...")
    subprocess.check_call(["git", "add", "."], cwd=cwd)
    subprocess.check_call(
        [
            "git",
            "commit",
            "--verify" if verify else "--no-verify",
            "-am",
            message,
        ],
        cwd=cwd,
    )


def git_checkout(branch_name, remote=False, mode="switch", cwd="./"):
    if mode == "switch":
        logger.info(f"Switch to branch {branch_name} ...")
        try:
            subprocess.check_call(["git", "show-branch", branch_name], cwd=cwd)
            # Branch exists
            subprocess.check_call(["git", "checkout", branch_name], cwd=cwd)
        except Exception:
            logger.debug(f"Branch {branch_name} doesn't exist. Create it ...")
            subprocess.check_call(
                ["git", "checkout", "--track", f"{remote}/{branch_name}"], cwd=cwd
            )
    elif mode == "create":
        logger.info(
            f"Create a new branch {branch_name}" " based on the current one ..."
        )
        subprocess.check_call(["git", "checkout", "-b", branch_name], cwd=cwd)
    elif mode == "create_orphan":
        logger.info(f"Create a new orphan branch {branch_name} ...")
        subprocess.check_call(["git", "checkout", "--orphan", branch_name], cwd=cwd)
        subprocess.check_call(["git", "reset", "--hard"], cwd=cwd)


def git_fetch(remote, branch_name, cwd="./"):
    logger.info(f"Fetch '{remote} {branch_name}' ...")
    commands = ["git", "fetch", remote, branch_name]
    logger.debug(f"Calling: {commands}")
    subprocess.check_call(commands, cwd=cwd)


def git_pull_rebase(remote, branch_name, cwd="./"):
    logger.info(f"Pull and rebase against '{remote} {branch_name}' ...")
    commands = ["git", "pull", remote, branch_name, "--rebase"]
    logger.debug(f"Calling: {commands}")
    subprocess.check_call(commands, cwd=cwd)


def git_push(remote, branch_name, force_push=False, cwd="./"):
    logger.info(f"Force Push on '{remote} {branch_name}' ...")
    commands = ["git", "push", remote, branch_name]
    if force_push:
        commands.append("--force")
    logger.debug(f"Calling: {commands}")
    subprocess.check_call(commands, cwd=cwd)


def git_switch(repo, remote, version, do_pull=False, cwd="./"):
    if repo.active_branch.name != version:
        try:
            logger.info(f"Switching on {version} ...")
            subprocess.check_call(["git", "switch", version], cwd=cwd)
        except Exception:
            logger.error(f"The branch '{version}' doesn't exist.")
            return

    try:
        logger.info(f"Pulling last revision of {version} ...")
        subprocess.check_call(["git", "pull", remote, version], cwd=cwd)
    except Exception:
        logger.error(f"Pulling '{remote}/{version}' failed.")
        return


def execute(command, cwd="./"):
    logger.debug(f"run command\n{' '.join(command)}")
    stdout = subprocess.run(command, cwd=cwd, check=True, stdout=subprocess.PIPE)
    logger.debug(stdout)


def replace_expression(file, search_expression, replace_expression):
    for line in fileinput.input(file, inplace=1):
        if search_expression in line:
            if replace_expression:
                line = line.replace(search_expression, replace_expression)
            else:
                continue
        sys.stdout.write(line)


def ensure_line(file_path, required_line):
    file = file_path.open(mode="r")
    lines = file.readlines()
    file.close()
    for line in lines:
        if line == required_line:
            return

    file = file_path.open(mode="a")
    file.write(str(f"{required_line}\n"))
    file.close()



import configparser
from pathlib import Path

import click
import yaml
from loguru import logger

from odoo_tools_grap.cli.click_options import option_repo_config_file


@click.command()
@click.option(
    "-i",
    "--input-files",
    multiple=True,
    type=click.Path(
        exists=True,
        dir_okay=False,
        resolve_path=True,
    ),
    help="Template odoo Config File(s)",
)
@click.option(
    "-o",
    "--output-file",
    type=click.Path(
        dir_okay=False,
        resolve_path=True,
    ),
    default="./odoo.cfg",
    help="Generated Odoo config file",
)
@option_repo_config_file
@click.pass_context
def generate(ctx, input_files, output_file, config_repo_file):
    """Generate a odoo.cfg file, based on:
    - one or more 'config.cfg ;
    - one repos.yml file, to generate addons_path.
    """
    input_files = [Path(x) for x in input_files]
    output_file = Path(output_file)
    config_repo_file = Path(config_repo_file)

    # Read Input Files
    parser = configparser.ConfigParser()
    parser.read(input_files)

    # Compute Addons path
    stream = open(config_repo_file)
    data = yaml.safe_load(stream)

    addons_path = []
    for key in data.keys():
        path = config_repo_file.cwd() / key
        if path.name == "odoo":
            # Append subfolder addons path
            addons_path.append(str(path / "addons"))
        else:
            addons_path.append(str(path))

    parser.set("options", "addons_path", "\n" + ",\n".join(addons_path))

    parser.write(open(output_file, "w"))
    logger.info("%s has been generated or updated." % (output_file))



import logging
import sys

import click
from click_loglevel import LogLevel
from loguru import logger

import odoo_tools_grap
from odoo_tools_grap.cli.cli_create_branch import create_branch
from odoo_tools_grap.cli.cli_diff import diff
from odoo_tools_grap.cli.cli_generate import generate
from odoo_tools_grap.cli.cli_migrate import migrate
from odoo_tools_grap.cli.cli_rebase import rebase


@click.group()
@click.version_option(odoo_tools_grap.__version__, "-v", "--version")
@click.option("-l", "--log-level", type=LogLevel(), default=logging.INFO)
@click.pass_context
def main(ctx, log_level):
    """
    Provides a command set to perform odoo environment manipulations.
    """
    logger.remove()
    logger.add(sys.stderr, level=log_level)


main.add_command(create_branch)
main.add_command(diff)
main.add_command(generate)
main.add_command(migrate)
main.add_command(rebase)



from pathlib import Path

import click
import yaml
from git import Repo
from loguru import logger

from odoo_tools_grap.cli.click_options import option_repo_config_file


@click.command()
@option_repo_config_file
@click.pass_context
def diff(ctx, config_repo_file):
    """parse a repo.yml config file, and for each repo
    check if repo is clean, on a correct target branch, etc...
    """
    config_repo_file = Path(config_repo_file)

    # Compute Addons path
    stream = open(config_repo_file)
    data = yaml.safe_load(stream)

    for addons_path, repo_data in data.items():
        path = config_repo_file.cwd() / addons_path
        logger.debug(f"Scanning repository '{addons_path}' ...")

        if not path.exists():
            logger.error(
                f"[NOT FOUND] The folder '{addons_path}' has not been found."
                " Maybe you could run:\n"
                f" gitaggregate -c repos.yml -d {addons_path}"
            )
            continue

        repo = Repo(str(path))

        # check if we are in the target repository
        current_branch = repo.active_branch.name
        target = repo_data.get("target", False)
        if target:
            target_branch = target.split(" ")[1]
        else:
            target_branch = "_git_aggregated"
        if current_branch != target_branch:
            logger.warning(
                f"[BAD BRANCH] {addons_path} is on {current_branch}."
                f"(Should be on {target_branch})"
            )

        # check local changes
        if repo.is_dirty():
            logger.warning(
                f"[LOCAL CHANGES] {addons_path} has"
                f" {len(repo.head.commit.diff(None))} local changes."
            )
        if repo.untracked_files:
            logger.warning(
                f"[UNTRACKED] {addons_path} has"
                f" {len(repo.untracked_files)} untracked files."
            )
